define(['exports', 'path', 'cosmiconfig', 'micromatch', 'deepmerge', '@structured-types/api/types', './types.amd', 'fs', 'hosted-git-info', '@hutson/parse-repository-url', 'js-string-escape'], (function (exports, path, cosmiconfig, micromatch, deepmerge, types$1, types, fs, hostedGitInfo, parseRepositoryURL, jsStringEscape) { 'use strict';

  function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

  function _interopNamespace(e) {
    if (e && e.__esModule) return e;
    var n = Object.create(null);
    if (e) {
      Object.keys(e).forEach(function (k) {
        if (k !== 'default') {
          var d = Object.getOwnPropertyDescriptor(e, k);
          Object.defineProperty(n, k, d.get ? d : {
            enumerable: true,
            get: function () { return e[k]; }
          });
        }
      });
    }
    n["default"] = e;
    return Object.freeze(n);
  }

  var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
  var path__namespace = /*#__PURE__*/_interopNamespace(path);
  var micromatch__default = /*#__PURE__*/_interopDefaultLegacy(micromatch);
  var deepmerge__default = /*#__PURE__*/_interopDefaultLegacy(deepmerge);
  var fs__namespace = /*#__PURE__*/_interopNamespace(fs);
  var hostedGitInfo__default = /*#__PURE__*/_interopDefaultLegacy(hostedGitInfo);
  var parseRepositoryURL__default = /*#__PURE__*/_interopDefaultLegacy(parseRepositoryURL);
  var jsStringEscape__default = /*#__PURE__*/_interopDefaultLegacy(jsStringEscape);

  const textNode = (value) => ({ kind: types.NodeKind.Text, value });

  const paragraphNode = (value) => ({
      kind: types.NodeKind.Paragraph,
      children: typeof value === 'string' ? [textNode(value)] : value,
  });

  const codeNode = (value) => ({
      kind: types.NodeKind.Code,
      value,
  });

  const boldNode = (value) => ({
      kind: types.NodeKind.Bold,
      children: typeof value === 'string' ? [textNode(value)] : value,
  });

  const defaultSections = {
      title: {},
      type: {},
      description: {},
      location: {},
      extends: { title: 'extends' },
      props: {
          title: (prop) => types$1.isFunctionProp(prop) ? 'parameters' : 'properties',
      },
      examples: {
          title: (prop) => prop.examples && prop.examples.length > 1 ? 'examples' : 'example',
      },
  };

  const defaultColumns = {
      name: { title: 'Name' },
      type: { title: 'Type' },
      parents: { title: 'Parent' },
      default: {
          title: (prop) => (types$1.isEnumProp(prop) ? 'Value' : 'Default'),
      },
      description: { title: 'Description' },
  };

  const headingNode = (value, depth) => ({
      kind: types.NodeKind.Heading,
      depth,
      children: typeof value === 'string' ? [textNode(value)] : value,
  });

  const titleSection = (prop) => {
      return prop.name ? [headingNode(prop.name, 2)] : undefined;
  };

  const descriptionSection = (prop) => {
      return prop.description
          ? [
              paragraphNode(prop.description.split('\n').map((d) => ({
                  kind: types.NodeKind.Text,
                  value: d,
              }))),
          ]
          : undefined;
  };

  const inlineCodeNode = (value) => value
      ? {
          kind: types.NodeKind.InlineCode,
          value,
      }
      : {
          kind: types.NodeKind.Text,
          value,
      };

  const linkNode = (children, url, loc) => {
      const prop = {
          kind: types.NodeKind.Link,
          children,
      };
      if (url) {
          prop.url = url;
      }
      if (loc) {
          prop.loc = loc;
      }
      return prop;
  };

  class PropLinks {
      constructor() {
          this.props = {};
      }
      init(props) {
          this.props = props;
      }
      getPropLink(key) {
          const nameParts = key.split('.');
          return this.props[nameParts[nameParts.length - 1]];
      }
      propLink(prop) {
          var _a;
          const typeText = inlineCodeNode(prop.name);
          if (typeof prop.name === 'string') {
              const link = this.getPropLink(prop.name);
              return linkNode([typeText], link ? `#${(_a = link.name) === null || _a === void 0 ? void 0 : _a.toLowerCase()}` : undefined, prop.loc);
          }
          return typeText;
      }
  }

  const typeSection = (prop, config) => {
      if (prop.kind) {
          return [
              boldNode([
                  inlineCodeNode(`${prop.extension ? `${prop.extension} ` : ''}${types$1.PropKind[prop.kind].toLowerCase()}`),
              ]),
          ];
      }
      else if (typeof prop.type === 'string') {
          return [
              headingNode([config.propLinks.propLink({ name: prop.type, loc: prop.loc })], 3),
          ];
      }
      return undefined;
  };

  const traverseFolder = (filePath, levels = 10, fileName = 'package.json') => {
      if (!fs__namespace.existsSync(filePath)) {
          return null;
      }
      const files = fs__namespace.readdirSync(filePath);
      if (levels === 0) {
          return null;
      }
      const pckg = files.find((file) => file === fileName);
      if (pckg) {
          return path__default["default"].resolve(filePath, pckg);
      }
      return traverseFolder(path__default["default"].resolve(filePath, '..'), levels - 1, fileName);
  };
  /**
   * Retrieves the repo path from the project's package.json file.
   * @param filePath file path to start the search for a package.json
   */
  const getRepoPath = (filePath) => {
      const packageFileName = traverseFolder(path__default["default"].dirname(filePath));
      const result = {};
      if (packageFileName) {
          const packageJSON = JSON.parse(fs__namespace.readFileSync(packageFileName, 'utf8'));
          result.packageName = packageJSON.name;
          result.relativePath = path__default["default"].relative(path__default["default"].dirname(packageFileName), filePath);
          const repositoryURL = typeof packageJSON.repository === 'string'
              ? packageJSON.repository
              : packageJSON.repository && packageJSON.repository.url;
          if (repositoryURL) {
              const templates = hostedGitInfo__default["default"].fromUrl(repositoryURL) ||
                  parseRepositoryURL__default["default"](repositoryURL);
              result.relativePath = `${packageJSON.repository && packageJSON.repository.directory
                ? `${packageJSON.repository.directory}/`
                : ''}${path__default["default"].relative(path__default["default"].dirname(packageFileName), filePath)}`;
              const fillTemplate = (template) => {
                  return template
                      .replace('auth@', templates.auth || '')
                      .replace('{domain}', templates.domain || '')
                      .replace('{user}', templates.user || '')
                      .replace('{project}', templates.project || '')
                      .replace('{#fragment}', '')
                      .replace('{path}', result.relativePath)
                      .replace('{treepath}', templates.treepath || '')
                      .replace('{/tree/committish}', `/${templates.treepath}/${templates.committish || 'master'}`)
                      .replace('{committish}', templates.committish || 'master');
              };
              result.repo = fillTemplate(templates.browsefiletemplate);
          }
      }
      return result;
  };

  class PropRepos {
      constructor() {
          this.repos = {};
      }
      getRepo(filePath) {
          if (!this.repos[filePath]) {
              this.repos[filePath] = getRepoPath(path__default["default"].resolve(filePath));
          }
          return this.repos[filePath];
      }
  }

  const emphasisNode = (value) => ({
      kind: types.NodeKind.Emphasis,
      children: typeof value === 'string' ? [textNode(value)] : value,
  });

  const locationSection = (prop, repos) => {
      const { filePath, loc } = prop.loc || {};
      if (filePath) {
          const repo = repos.getRepo(filePath);
          if (repo) {
              const { repo: repoName, relativePath, packageName } = repo;
              const fileLocation = repoName || filePath;
              if (fileLocation) {
                  const { line } = (loc === null || loc === void 0 ? void 0 : loc.start) || {};
                  const sourceLocation = filePath.includes('node_modules')
                      ? fileLocation
                      : `${fileLocation}${line ? `#L${line}` : ''}`;
                  return [
                      paragraphNode([
                          emphasisNode([
                              textNode('defined in '),
                              linkNode([
                                  textNode(packageName
                                      ? `${packageName}/${relativePath}`
                                      : sourceLocation),
                              ], sourceLocation, prop.loc),
                          ]),
                      ]),
                  ];
              }
          }
      }
      return undefined;
  };

  const extendsSection = (prop, config) => {
      var _a;
      if (types$1.isClassLikeProp(prop) && ((_a = prop.extends) === null || _a === void 0 ? void 0 : _a.length)) {
          const extendsList = prop.extends.reduce((acc, extendsProp, idx) => {
              const result = [
                  config.propLinks.propLink(extendsProp),
              ];
              if (prop.extends && idx < prop.extends.length - 1) {
                  result.push(textNode(', '));
              }
              return [...acc, ...result];
          }, []);
          return extendsList;
      }
      return undefined;
  };

  const collapsibleNode = (children, summary) => ({
      kind: types.NodeKind.Collapsible,
      summary,
      children,
  });

  const blockNode = (children) => ({
      kind: types.NodeKind.Block,
      children,
  });

  const shortPropType = (prop, config) => {
      if (typeof prop.type === 'string') {
          return config.propLinks.propLink({
              name: prop.type,
              loc: prop.loc,
          });
      }
      if (prop.kind) {
          return inlineCodeNode(`${types$1.PropKind[prop.kind].toLowerCase()}`);
      }
      return undefined;
  };

  const getPropValue = (prop) => {
      if (types$1.hasValue(prop) && prop.value !== undefined) {
          return types$1.isStringProp(prop)
              ? `"${jsStringEscape__default["default"](prop.value)}"`
              : prop.value.toString();
      }
      else if (types$1.isNullProp(prop)) {
          return 'null';
      }
      else if (types$1.isUndefinedProp(prop)) {
          return 'undefined';
      }
      else if (types$1.isVoidProp(prop)) {
          return 'void';
      }
      return undefined;
  };
  const unionPropNodes = (prop, config) => {
      const unionSeparator = ' | ';
      const enumSeparator = ', ';
      const separator = types$1.isUnionProp(prop) ? unionSeparator : enumSeparator;
      const properties = prop.properties;
      if (properties) {
          const propsList = properties.reduce((acc, p, idx) => {
              let propType = undefined;
              if (p.parent && p.name) {
                  propType = [
                      config.propLinks.propLink(p.parent),
                      textNode('.'),
                      config.propLinks.propLink({ name: p.name, loc: p.loc }),
                  ];
              }
              else {
                  const value = getPropValue(p);
                  if (typeof value === 'string') {
                      propType = [inlineCodeNode(value)];
                  }
                  else {
                      const pt = shortPropType(p, config);
                      if (pt) {
                          propType = [pt];
                      }
                  }
              }
              if (propType) {
                  const r = [...acc, ...propType];
                  if (idx < properties.length - 1) {
                      r.push(textNode(separator));
                  }
                  return r;
              }
              return acc;
          }, []);
          return propsList;
      }
      return [];
  };

  const arrayPropNodes = (prop, config) => {
      const properties = prop.properties;
      if (properties) {
          const elements = properties.reduce((acc, p, idx) => {
              const propType = shortPropType(p, config);
              if (propType) {
                  const result = [propType];
                  if (idx < properties.length - 1) {
                      result.push(textNode(', '));
                  }
                  return [...acc, ...result];
              }
              return acc;
          }, []);
          const multiProps = elements.length > 1;
          if (multiProps) {
              elements.splice(0, 0, textNode('('));
              elements.push(textNode(')'));
          }
          elements.push(textNode('[]'));
          return elements;
      }
      return [];
  };

  const tuplePropNodes = (prop, config) => {
      const properties = prop.properties;
      if (properties) {
          const nodes = [
              textNode('['),
              ...properties.reduce((acc, p, idx) => {
                  const propType = shortPropType(p, config);
                  if (propType) {
                      const result = [propType];
                      if (idx < properties.length - 1) {
                          result.push(textNode(', '));
                      }
                      return [...acc, ...result];
                  }
                  return acc;
              }, []),
              textNode(']'),
          ];
          return nodes;
      }
      return [];
  };

  const classPropNodes = (prop, config) => {
      const properties = prop.properties;
      const propName = typeof prop.type === 'string' ? prop.type : prop.name;
      if (properties) {
          const typeArguments = properties.reduce((acc, p) => {
              return [...acc, ...config.propTypes.extractNamedType(p)];
          }, []);
          return typeArguments;
      }
      else if (prop.generics) {
          const generics = prop.generics;
          const result = [];
          if (propName) {
              result.push(config.propLinks.propLink({
                  name: propName,
                  loc: prop.loc,
              }));
          }
          result.push(textNode('<'));
          generics.forEach((p, idx) => {
              const propType = shortPropType(p, config);
              if (propType) {
                  result.push(propType);
                  if (idx < generics.length - 1) {
                      result.push(textNode(', '));
                  }
              }
          });
          result.push(textNode('>'));
          return result;
      }
      else if (propName) {
          return [textNode(propName)];
      }
      return [];
  };

  const indexPropNodes = (prop, config) => {
      const propType = shortPropType(prop.index, config);
      const result = [];
      if (propType) {
          result.push(textNode('['), propType, textNode(']'));
      }
      if (prop.prop) {
          result.push(textNode(': '));
          result.push(...config.propTypes.extractType(prop.prop));
      }
      return [blockNode(result)];
  };

  const functionPropNodes = (prop, config) => {
      const result = [
          blockNode([boldNode('function'), textNode(' (')]),
      ];
      if (prop.parameters) {
          prop.parameters.forEach((p) => {
              result.push(...config.propTypes.extractNamedType(p));
          });
      }
      result.push(textNode(')'));
      if (prop.returns) {
          const propReturns = {
              name: 'returns',
              ...prop.returns,
          };
          result.push(textNode(' => '));
          result.push(...config.propTypes.extractType(propReturns));
      }
      else {
          result.push(textNode('=> void'));
      }
      return result;
  };

  class PropTypeNodes {
      constructor() {
          this.config = {};
      }
      init(config) {
          this.config = config;
      }
      extractNamedType(prop) {
          return this.extractType(prop, true);
      }
      extractType(prop, named = false) {
          var _a;
          const linkedProp = this.resolvedProp(prop);
          const typeNode = this.getType(linkedProp);
          let nameNode = undefined;
          if (named) {
              if (linkedProp.name && linkedProp.name !== linkedProp.type) {
                  nameNode = [inlineCodeNode(`${linkedProp.name}`)];
                  if (!linkedProp.optional) {
                      nameNode.push(textNode('*'));
                  }
              }
          }
          if (types$1.isClassLikeProp(linkedProp) &&
              !this.isLinkedProp(linkedProp) &&
              !!((_a = linkedProp.properties) === null || _a === void 0 ? void 0 : _a.length)) {
              if (named) {
                  return [this.collapsibleType(linkedProp, typeNode, nameNode)];
              }
              else {
                  const propType = shortPropType(linkedProp, this.config);
                  if (propType) {
                      return [this.collapsibleType(linkedProp, typeNode, [propType])];
                  }
              }
          }
          return nameNode
              ? [blockNode([...nameNode, textNode(': '), ...typeNode])]
              : typeNode;
      }
      resolvedProp(prop) {
          var _a;
          if (prop.parent) {
              const parent = this.config.propLinks.getPropLink(prop.parent.name);
              if (parent && types$1.isClassLikeProp(parent)) {
                  const p = (_a = parent.properties) === null || _a === void 0 ? void 0 : _a.find((p) => p.name === prop.name);
                  if (p) {
                      return p;
                  }
              }
          }
          return prop;
      }
      collapsibleType(prop, typeProp, nameNode) {
          var _a;
          const name = Array.isArray(nameNode)
              ? nameNode
              : nameNode
                  ? [nameNode]
                  : [
                      textNode(((_a = prop.properties) === null || _a === void 0 ? void 0 : _a.length)
                          ? `...${prop.properties.length} properties`
                          : 'expand'),
                  ];
          return collapsibleNode(typeProp, name);
      }
      isLinkedProp(prop) {
          const propName = typeof prop.type === 'string' ? prop.type : prop.name;
          if (propName) {
              const linkedProp = this.config.propLinks.getPropLink(propName);
              if (linkedProp && linkedProp !== prop) {
                  return true;
              }
          }
          return false;
      }
      getType(prop) {
          const propName = typeof prop.type === 'string' ? prop.type : prop.name;
          if (typeof propName === 'string') {
              const linkedProp = this.config.propLinks.getPropLink(propName);
              if (linkedProp) {
                  return [
                      this.config.propLinks.propLink({
                          name: propName,
                          loc: linkedProp.loc,
                      }),
                  ];
              }
          }
          return this.extractTypeNode(prop);
      }
      extractTypeNode(prop) {
          var _a;
          if (typeof prop.type === 'string' &&
              ((_a = this.config.options.collapsed) === null || _a === void 0 ? void 0 : _a.includes(prop.type))) {
              const propType = shortPropType(prop, this.config);
              if (propType) {
                  return [propType];
              }
          }
          else if (types$1.isUnionProp(prop) || types$1.isEnumProp(prop)) {
              return unionPropNodes(prop, this.config);
          }
          else if (types$1.isClassLikeProp(prop)) {
              return classPropNodes(prop, this.config);
          }
          else if (types$1.isArrayProp(prop)) {
              return arrayPropNodes(prop, this.config);
          }
          else if (types$1.isTupleProp(prop)) {
              return tuplePropNodes(prop, this.config);
          }
          else if (types$1.isIndexProp(prop)) {
              return indexPropNodes(prop, this.config);
          }
          else if (types$1.isFunctionProp(prop)) {
              return functionPropNodes(prop, this.config);
          }
          const propType = shortPropType(prop, this.config);
          if (propType) {
              return [propType];
          }
          return [];
      }
  }

  const tableCellNode = (value) => ({
      kind: types.NodeKind.TableCell,
      children: typeof value === 'string' ? [textNode(value)] : value,
  });

  const createPropsRow = ({ name, type, default: defaultValue, parents, description, prop }, columnOptions, visColumns) => {
      const getCell = (column) => {
          switch (column) {
              case 'name':
                  return name;
              case 'type':
                  return type;
              case 'default':
                  return [
                      inlineCodeNode(typeof defaultValue !== 'undefined' ? defaultValue.toString() : ''),
                  ];
              case 'description':
                  if (typeof description === 'string') {
                      const parts = description.split('`');
                      if (parts.length > 1) {
                          return parts.reduce((acc, text, idx) => {
                              if (idx % 2 === 0) {
                                  return [...acc, textNode(text)];
                              }
                              else {
                                  return [
                                      ...acc,
                                      textNode(' '),
                                      inlineCodeNode(text),
                                      textNode(' '),
                                  ];
                              }
                          }, []);
                      }
                      else {
                          return [textNode(description)];
                      }
                  }
                  else {
                      return [textNode('')];
                  }
              case 'parents':
                  return parents;
              default:
                  return undefined;
          }
      };
      const children = [];
      Object.keys(columnOptions).forEach((key) => {
          const name = key;
          const column = columnOptions[name];
          if (column && visColumns[name]) {
              let cell = undefined;
              if (typeof column.render === 'function') {
                  cell = column.render(name, prop);
              }
              if (!cell) {
                  const content = getCell(name);
                  cell = typeof content === 'string' ? textNode(content) : content;
              }
              if (cell) {
                  if (Array.isArray(cell)) {
                      children.push(tableCellNode(cell));
                  }
                  else {
                      children.push(tableCellNode([cell]));
                  }
              }
          }
      });
      return {
          kind: types.NodeKind.TableRow,
          children,
      };
  };
  const createPropsTable = (children, columnOptions, prop) => {
      var _a;
      const propsTable = [];
      let table = undefined;
      const visibleColumns = {
          name: false,
          default: false,
          description: false,
          parents: false,
          type: true,
      };
      const getTitle = (config, prop) => {
          let value = '';
          const { title } = config;
          if (title) {
              value = (typeof title === 'function' ? title(prop) : title) || '';
          }
          return tableCellNode([textNode(value)]);
      };
      if (children) {
          const columns = [];
          Object.keys(columnOptions).forEach((key) => {
              const name = key;
              const column = columnOptions[name];
              if (column && !column.hidden) {
                  visibleColumns[name] = children.some((item) => item[name] !== undefined);
                  if (visibleColumns[name]) {
                      columns.push(getTitle(column, prop));
                  }
              }
          });
          table = {
              kind: types.NodeKind.Table,
              children: [
                  {
                      kind: types.NodeKind.TableRow,
                      children: columns,
                  },
              ],
          };
          propsTable.push(table);
          (_a = table.children) === null || _a === void 0 ? void 0 : _a.push(...children.map((child) => createPropsRow(child, columnOptions, visibleColumns)));
      }
      return { propsTable: propsTable, table, visibleColumns };
  };

  const propTable = (prop, rows, config) => {
      let parentProp = undefined;
      const addParentProp = (prop) => {
          var _a, _b;
          if (prop.parent) {
              if (!parentProp) {
                  parentProp = {
                      name: '...props',
                      kind: types$1.PropKind.Enum,
                      properties: [
                          {
                              kind: types$1.PropKind.Type,
                              type: prop.parent.name,
                              loc: prop.parent.loc,
                          },
                      ],
                      optional: true,
                  };
              }
              else {
                  if (!((_a = parentProp.properties) === null || _a === void 0 ? void 0 : _a.find((p) => { var _a; return p.type === ((_a = prop.parent) === null || _a === void 0 ? void 0 : _a.name); }))) {
                      (_b = parentProp.properties) === null || _b === void 0 ? void 0 : _b.push({
                          kind: types$1.PropKind.Type,
                          type: prop.parent.name,
                          loc: prop.parent.loc,
                      });
                  }
              }
          }
      };
      const consolidatedProps = rows.filter((prop) => {
          var _a, _b;
          if (prop.parent) {
              if (config.options.skipInherited ||
                  ((_a = config.options.collapsed) === null || _a === void 0 ? void 0 : _a.includes(prop.parent.name))) {
                  addParentProp(prop);
                  return false;
              }
              if (config.options.collapsed) {
                  for (const collapsedProp of config.options.collapsed) {
                      const helperParent = config.helpers[collapsedProp];
                      if (helperParent && types$1.hasProperties(helperParent)) {
                          const helpProp = (_b = helperParent.properties) === null || _b === void 0 ? void 0 : _b.find((p) => { var _a, _b; return p.name === prop.name && ((_a = p.parent) === null || _a === void 0 ? void 0 : _a.name) === ((_b = prop.parent) === null || _b === void 0 ? void 0 : _b.name); });
                          if (helpProp) {
                              addParentProp({
                                  ...prop,
                                  parent: { name: collapsedProp, loc: helperParent.loc },
                              });
                              return false;
                          }
                      }
                  }
              }
          }
          return true;
      });
      const allProps = parentProp
          ? [...consolidatedProps, parentProp]
          : consolidatedProps;
      const items = allProps.map((prop) => {
          const nameProp = types$1.isIndexProp(prop) ? prop.index : prop;
          let name = nameProp.name
              ? prop.kind === types$1.PropKind.Rest
                  ? `...${nameProp.name}`
                  : nameProp.name
              : '';
          name = `${name}${!name || prop.optional ? '' : '*'}`;
          return configurePropItem({
              name: [config.propLinks.propLink({ name, loc: prop.loc })],
              parents: prop.parent
                  ? [config.propLinks.propLink(prop.parent)]
                  : undefined,
              type: config.propTypes.extractType(prop),
              description: prop.description,
              default: types$1.hasValue(prop)
                  ? types$1.isStringProp(prop) && typeof prop.value === 'string'
                      ? `"${prop.value}"`
                      : prop.value
                  : undefined,
              prop,
          }, config);
      });
      return createPropsTable(items, config.columns, prop);
  };
  const configurePropItem = (item, config) => {
      const enabledColumn = (name) => {
          var _a;
          return config.columns[name] && !((_a = config.columns[name]) === null || _a === void 0 ? void 0 : _a.hidden);
      };
      return {
          name: enabledColumn('name') ? item.name : undefined,
          parents: enabledColumn('parents') ? item.parents : undefined,
          type: enabledColumn('type') ? item.type : undefined,
          default: enabledColumn('default') ? item.default : undefined,
          description: enabledColumn('description') ? item.description : undefined,
          prop: item.prop,
      };
  };

  const propFunction = (prop, config) => {
      if (prop.parameters) {
          const { propsTable, table, visibleColumns } = propTable(prop, prop.parameters, config);
          if (table &&
              table.children &&
              prop.returns &&
              prop.returns.kind !== types$1.PropKind.Void) {
              table.children.push(createPropsRow(configurePropItem({
                  name: [inlineCodeNode('returns')],
                  parents: prop.returns.parent
                      ? [config.propLinks.propLink(prop.returns.parent)]
                      : undefined,
                  type: config.propTypes.extractType(prop.returns),
                  description: prop.returns.description,
                  default: undefined,
                  prop,
              }, config), config.columns, visibleColumns));
          }
          return propsTable;
      }
      return undefined;
  };

  class PropsToDocumentation {
      constructor() {
          this.sections = defaultSections;
          this.config = {
              columns: defaultColumns,
              propLinks: new PropLinks(),
              helpers: {},
              options: {},
              propTypes: new PropTypeNodes(),
          };
          this.repos = new PropRepos();
      }
      extractClassLike(prop) {
          var _a;
          if (prop.name && ((_a = prop.properties) === null || _a === void 0 ? void 0 : _a.length)) {
              const result = [];
              if (types$1.isUnionProp(prop)) {
                  result.push(...unionPropNodes(prop, this.config));
              }
              else if (types$1.hasProperties(prop)) {
                  const { propsTable } = propTable(prop, prop.properties, this.config);
                  result.push(...propsTable);
              }
              return result;
          }
          return undefined;
      }
      getPropsTable(prop) {
          if (types$1.isFunctionProp(prop)) {
              return propFunction(prop, this.config);
          }
          else if (types$1.hasProperties(prop)) {
              return this.extractClassLike(prop);
          }
          return undefined;
      }
      getExamples(prop) {
          if (prop.examples) {
              const result = [];
              const codeExamples = prop.examples.filter((e) => e.content);
              codeExamples.forEach((example) => {
                  if (example.content) {
                      result.push(paragraphNode([codeNode(example.content)]));
                  }
              });
              return result;
          }
          return undefined;
      }
      getSection(prop, name) {
          switch (name) {
              case 'title':
                  return titleSection(prop);
              case 'description':
                  return descriptionSection(prop);
              case 'type':
                  return typeSection(prop, this.config);
              case 'location':
                  return locationSection(prop, this.repos);
              case 'props':
                  return this.getPropsTable(prop);
              case 'extends':
                  return extendsSection(prop, this.config);
              case 'examples':
                  return this.getExamples(prop);
              default:
                  return undefined;
          }
      }
      generateSection(prop, name) {
          const config = this.sections[name];
          if (config) {
              const { title, render } = config;
              const children = typeof render === 'function'
                  ? render(prop)
                  : this.getSection(prop, name);
              if (children) {
                  const result = [];
                  if (title) {
                      const value = typeof title === 'function' ? title(prop) : title;
                      if (value) {
                          result.push(paragraphNode([boldNode([textNode(value)])]));
                      }
                  }
                  result.push(...children);
                  return result;
              }
          }
          return undefined;
      }
      generateNodes(prop) {
          const result = [];
          Object.keys(this.sections).forEach((name) => {
              const config = this.sections[name];
              if (!(config === null || config === void 0 ? void 0 : config.hidden)) {
                  const section = this.generateSection(prop, name);
                  if (section) {
                      result.push(...section);
                  }
              }
          });
          return result;
      }
      extract(props, options = {}) {
          const result = [];
          const { extensions, visible, sections = {}, columns = {} } = options;
          this.config.options = options;
          if (sections) {
              this.sections = Array.isArray(sections)
                  ? sections.reduce((acc, name) => {
                      return { ...acc, [name]: defaultSections[name] };
                  }, {})
                  : deepmerge__default["default"](defaultSections, sections);
          }
          if (columns) {
              this.config.columns = Array.isArray(columns)
                  ? columns.reduce((acc, name) => {
                      return { ...acc, [name]: defaultColumns[name] };
                  }, {})
                  : deepmerge__default["default"](defaultColumns, columns);
          }
          const filterProps = (prop) => {
              if (Array.isArray(extensions)) {
                  if (typeof prop.extension === 'string' &&
                      extensions.includes(prop.extension)) {
                      return prop;
                  }
                  return undefined;
              }
              if (Array.isArray(visible)) {
                  if (typeof prop.name === 'string' && visible.includes(prop.name)) {
                      return prop;
                  }
                  return undefined;
              }
              return prop;
          };
          const toplevelProps = {};
          Object.keys(props).forEach((key) => {
              if (key !== '__helpers' && key !== '__diagnostics') {
                  const prop = filterProps(props[key]);
                  if (prop) {
                      toplevelProps[key] = prop;
                  }
              }
          });
          this.config.helpers = props.__helpers || {};
          Object.keys(this.config.helpers).forEach((key) => {
              const prop = filterProps(this.config.helpers[key]);
              if (prop) {
                  toplevelProps[key] = prop;
              }
          });
          this.config.propLinks.init(toplevelProps);
          this.config.propTypes.init(this.config);
          Object.values(toplevelProps).forEach((prop) => {
              const nodes = this.generateNodes(prop);
              result.push(...nodes);
          });
          return result;
      }
  }

  /**
   * Creates a list of api documentation nodes
   * @param props properties parsed from `structured-types/api`
   * @param options page generation options
   * @returns a list of documentation nodes
   */
  const propsToDocumentation = (props, options) => {
      const extractor = new PropsToDocumentation();
      return extractor.extract(props, options);
  };
  /**
   * deep merge of documentation configurations, by replacing array members (ie sections, columns)
   */
  const mergeConfig = (dest, src) => deepmerge__default["default"](dest, src, {
      arrayMerge: (dest, src) => src,
  });
  /**
   * Read the api-docs configuration file
   * @param fileName the file that is being analyzed, will be used the starting folder to search for configuration files.
   * @param configFileName pass directly the configuration file name
   * @param elementId an optional element id. If not, the elements will be micromatch'ed based on relative path from configuration
   * @returns page generation options from the config file
   */
  const apiDocsConfig = (fileName, configFileName, elementId) => {
      const configExplorer = cosmiconfig.cosmiconfigSync('api-docs');
      let results;
      if (configFileName) {
          results = configExplorer.load(configFileName);
      }
      else {
          const searchPath = path__namespace.dirname(fileName);
          results = configExplorer.search(searchPath);
      }
      if (results) {
          const { filepath, config } = results;
          if (filepath && config) {
              const { elements, ...rest } = config;
              results.config = rest;
              if (elements) {
                  const id = elementId || path__namespace.relative(path__namespace.dirname(filepath), fileName);
                  const onlyName = path__namespace.basename(fileName);
                  Object.keys(elements).forEach((key) => {
                      if (onlyName === key || micromatch__default["default"].isMatch(id, key)) {
                          results.config = mergeConfig(results.config, elements[key]);
                      }
                  });
              }
          }
      }
      return results;
  };

  Object.defineProperty(exports, 'NodeKind', {
    enumerable: true,
    get: function () { return types.NodeKind; }
  });
  exports.isBlockNode = types.isBlockNode;
  exports.isBoldNode = types.isBoldNode;
  exports.isCodeNode = types.isCodeNode;
  exports.isCollapsibleNode = types.isCollapsibleNode;
  exports.isEmphasisNode = types.isEmphasisNode;
  exports.isHeadingNode = types.isHeadingNode;
  exports.isInlineCodeNode = types.isInlineCodeNode;
  exports.isLinkNode = types.isLinkNode;
  exports.isNodeWithChildren = types.isNodeWithChildren;
  exports.isNodeWithValue = types.isNodeWithValue;
  exports.isParagraphNode = types.isParagraphNode;
  exports.isTableCellNode = types.isTableCellNode;
  exports.isTableNode = types.isTableNode;
  exports.isTableRowNode = types.isTableRowNode;
  exports.isTextNode = types.isTextNode;
  exports.apiDocsConfig = apiDocsConfig;
  exports.getRepoPath = getRepoPath;
  exports.mergeConfig = mergeConfig;
  exports.propsToDocumentation = propsToDocumentation;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
